@@ -1,16 +1,18 @@
 #!/usr/bin/env python3
 
-from sys import exit
-from os import path
 from argparse import ArgumentParser
-from json import dump as json_dump, load as json_load
+from json import dump as json_dump
 import base64
 import binascii
 from math import fabs
 from re import compile as re_compile
 from datetime import date, datetime, timedelta
 from commandparse import Command
-from Cryptodome.Hash import MD4
+from time import sleep
+
+from Cryptodome.Cipher import AES
+from Cryptodome.Hash import MD4, SHA1
+from Cryptodome.Protocol.KDF import PBKDF2
 
 from pyasn1.error import PyAsn1UnicodeDecodeError
 
@@ -55,6 +57,8 @@ class Ldeep(Command):
 					print(record["dc"])
 				elif "domain" in record["objectClass"]:
 					print(record["dn"])
+				elif "pKIEnrollmentService" in record["objectClass"]:
+					print(record["dNSHostName"])
 
 	def __display_json(self, records, default):
 		json_dump(records, sys.stdout, ensure_ascii=False, default=default, sort_keys=True, indent=2)
@@ -178,17 +182,46 @@ class Ldeep(Command):
 			print("Can't retrieve gmsa, TLS needed")
 			return
 		entries = self.engine.query("(ObjectClass=msDS-GroupManagedServiceAccount)", attributes)
+
+		constants = b'\x6b\x65\x72\x62\x65\x72\x6f\x73\x7b\x9b\x5b\x2b\x93\x13\x2b\x93\x5c\x9b\xdc\xda\xd9\x5c\x98\x99\xc4\xca\xe4\xde\xe6\xd6\xca\xe4'
+		iv = b'\x00' * 16
+
 		for entry in entries:
 			try:
 				sam = entry['sAMAccountName']
 				data = entry['msDS-ManagedPassword']
 				blob = MSDS_MANAGEDPASSWORD_BLOB()
 				blob.fromString(data)
-				hash = MD4.new ()
-				hash.update (blob['CurrentPassword'][:-2])
-				passwd = binascii.hexlify(hash.digest()).decode("utf-8")
-				userpass = sam + ':::' + passwd
-				print(userpass)
+				password = blob['CurrentPassword'][:-2]
+
+				# Compute NT hash
+				hash = MD4.new()
+				hash.update(password)
+				nthash = hash.hexdigest()
+
+				# Quick and dirty way to get the FQDN of the account's domain
+				dc_list = []
+				for s in entry['dn'].split(','):
+					if s.startswith('DC='):
+						dc_list.append(s[3:])
+
+				domain_fqdn = '.'.join(dc_list)
+				salt = f'{domain_fqdn.upper()}host{sam[:-1].lower()}.{domain_fqdn.lower()}'
+				encryption_key = PBKDF2(password.decode('utf-16-le', 'replace').encode(), salt.encode(), 32, count=4096, hmac_hash_module=SHA1)
+
+				# Compute AES keys
+				cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
+				first_part = cipher.encrypt(constants)
+				cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
+				second_part = cipher.encrypt(first_part)
+				aes256_key = first_part[:16] + second_part[:16]
+
+				cipher = AES.new(encryption_key[:16], AES.MODE_CBC, iv)
+				aes128_key = cipher.encrypt(constants[:16])
+
+				print(f'{sam}:nt:{nthash}')
+				print(f'{sam}:aes128-cts-hmac-sha1-96:{aes128_key.hex()}')
+				print(f'{sam}:aes256-cts-hmac-sha1-96:{aes256_key.hex()}')
 			except:
 				print(sam)
 
@@ -203,12 +236,39 @@ class Ldeep(Command):
 			"minPwdAge": (86400, "days"),
 			"forceLogoff": (60, "mins")
 		}
-		FIELDS_TO_PRINT = ["dc", "distinguishedName", "lockOutObservationWindow", "lockoutDuration", "lockoutThreshold", "maxPwdAge", "minPwdAge", "minPwdLength", "pwdHistoryLength", "pwdProperties"]
+
+		FOREST_LEVELS = {
+			7:"Windows Server 2016",
+			6:"Windows Server 2012 R2",
+			5:"Windows Server 2012",
+			4:"Windows Server 2008 R2",
+			3:"Windows Server 2008",
+			2:"Windows Server 2003",
+			1:"Windows Server 2003 operating system through Windows Server 2016",
+			0:"Windows 2000 Server operating system through Windows Server 2008 operating system"
+		}
+
+		FIELDS_TO_PRINT = [
+			"dc",
+			"distinguishedName",
+			"lockOutObservationWindow",
+			"lockoutDuration",
+			"lockoutThreshold",
+			"maxPwdAge",
+			"minPwdAge",
+			"minPwdLength",
+			"pwdHistoryLength",
+			"pwdProperties",
+			"ms-DS-MachineAccountQuota",
+			"msDS-Behavior-Version"]
+
 		policy = self.engine.query(self.engine.DOMAIN_INFO_FILTER())
 		if policy:
 			policy = policy[0]
 			for field in FIELDS_TO_PRINT:
-				val = policy[field]
+				val = policy.get(field, None)
+				if val is None:
+					continue
 
 				if field == "lockOutObservationWindow" and isinstance(val, timedelta):
 					val = int(val.total_seconds()) / 60
@@ -216,6 +276,8 @@ class Ldeep(Command):
 					val = int((fabs(float(val)) / 10**7) / FILETIME_TIMESTAMP_FIELDS[field][0])
 				if field in FILETIME_TIMESTAMP_FIELDS.keys():
 					val = "%d %s" % (val, FILETIME_TIMESTAMP_FIELDS[field][1])
+				if field == "msDS-Behavior-Version" and isinstance(val, int):
+					val = "%s" % (FOREST_LEVELS[policy[field]])
 
 				print("%s: %s" % (field, val))
 
@@ -256,6 +318,7 @@ class Ldeep(Command):
 			"msDS-MaximumPasswordAge": (86400, "days"),
 			"msDS-LockoutDuration": (60, "mins")
 		}
+
 		FIELDS_TO_PRINT = [
 			"cn",
 			"msDS-PasswordReversibleEncryptionEnabled",
@@ -268,12 +331,20 @@ class Ldeep(Command):
 			"msDS-LockoutThreshold",
 			"msDS-MinimumPasswordAge",
 			"msDS-MaximumPasswordAge",
+			"msDS-PSOAppliesTo",
 		]
+
 		psos = self.engine.query(self.engine.PSO_INFO_FILTER())
 		for policy in psos:
 			for field in FIELDS_TO_PRINT:
-				if isinstance(policy[field], list):
-					val = policy[field][0]
+				val = policy.get(field, None)
+				if not val:
+					continue
+				if isinstance(val, list):
+					targets = []
+					for target in val:
+						targets.append(target)
+					val = " | ".join(targets)
 				else:
 					val = policy[field]
 
@@ -287,7 +358,18 @@ class Ldeep(Command):
 		List the domain's trust relationships.
 		"""
 		results = self.engine.query(self.engine.TRUSTS_INFO_FILTER())
-		FIELDS_TO_PRINT = ["dn", "cn", "securityIdentifier", "name", "trustDirection", "trustPartner", "trustType", "trustAttributes", "flatName"]
+		FIELDS_TO_PRINT = [
+			"dn",
+			"cn",
+			"securityIdentifier",
+			"name",
+			"trustDirection",
+			"trustPartner",
+			"trustType",
+			"trustAttributes",
+			"flatName"
+		]
+
 		for result in results:
 			for field in FIELDS_TO_PRINT:
 				if field in result:
@@ -334,6 +416,76 @@ class Ldeep(Command):
 			verbose
 		)
 
+	def list_pkis(self, kwargs):
+		"""
+		List pkis.
+		Arguments:
+			@verbose:bool
+				Results will contain full information
+		"""
+		verbose  = kwargs.get("verbose", False)
+
+		if verbose:
+			attributes = self.engine.all_attributes()
+		else:
+			attributes = ["dNSHostName", "objectClass"]
+
+		self.display(
+			self.engine.query(self.engine.PKI_FILTER(),
+				attributes,
+				base=','.join(["CN=Enrollment Services,CN=Public Key Services,CN=Services,CN=CONFIGURATION", self.engine.base_dn]),
+				),
+				verbose
+			)
+
+	def list_subnets(self, kwargs):
+		"""
+		List sites and associated subnets.
+		Arguments:
+			@verbose:bool
+				Results will contain full information
+		"""
+		verbose = kwargs.get("verbose", False)
+
+		if not verbose:
+			attributes = ["distinguishedName", "name", "description"]
+		else:
+			attributes = ALL
+
+		if verbose:
+			self.display(
+				self.engine.query(
+					self.engine.SITES_FILTER(),
+					attributes, base=','.join(["CN=Configuration", self.engine.base_dn])
+				),
+				verbose
+			)
+		else:
+			entries = self.engine.query(self.engine.SITES_FILTER(), attributes, base=','.join(["CN=Configuration", self.engine.base_dn]))
+
+			site_dn = ""
+			site_name = ""
+			site_description = ""
+			subnet_dn = ""
+			subnet_name = ""
+			subnet_description = ""
+			for entry in entries:
+				site_dn = entry["distinguishedName"] if entry["distinguishedName"] else ""
+				site_name = entry["name"] if entry["name"] else ""
+				site_description = entry["description"][0] if entry["description"] else ""
+				subnet_entries = self.engine.query(self.engine.SUBNET_FILTER(site_dn), attributes, base=','.join(["CN=Sites,CN=Configuration", self.engine.base_dn]))
+				for subnet in subnet_entries:
+					subnet_dn = subnet["distinguishedName"] if subnet["distinguishedName"] else ""
+					subnet_name = subnet["name"] if subnet["name"] else ""
+					subnet_description = subnet["description"][0] if subnet["description"] else ""
+
+				output = "Site: {}".format(site_name)
+				output += " | Subnet: {}".format(subnet_name) if subnet_name else ""
+				output += " | Site description: {}".format(site_description) if site_description else ""
+				output += " | Subnet description: {}".format(subnet_description) if subnet_description else ""
+				print(output)
+
+
 	# GETTERS #
 
 	def get_zone(self, kwargs):
@@ -568,6 +720,32 @@ class Ldeep(Command):
 					getattr(self, method)(kwargs)
 					kwargs["verbose"] = False
 
+	def misc_enum_users(self, kwargs):
+		"""
+		Anonymously enumerate users with LDAP pings.
+
+		Arguments:
+			#file:string
+				File containing a list of usernames to try.
+			@delay:int = 0
+				Delay in milliseconds between each try.
+		"""
+
+		# LDAP pings can only be used with an anonymous bind
+		if self.engine.ldap.authentication != 'ANONYMOUS':
+			error('The enum_users feature can only be used with an anonymous bind (-a option)')
+
+		file = kwargs["file"]
+		delay = kwargs["delay"]
+		with open(file, 'r') as f:
+			while True:
+				l = f.readline()[:-1]
+				if not l:
+					break
+				if self.engine.user_exists(l):
+					print(l)
+				sleep(delay / 1000)
+
 	# ACTION #
 
 	def action_unlock(self, kwargs):
@@ -608,6 +786,42 @@ class Ldeep(Command):
 		else:
 			error("Unable to change {username}'s password, check privileges or try with ldaps://".format(username=user))
 
+	def action_add_to_group(self, kwargs):
+		"""
+		Add `user` to `group`.
+
+		Arguments:
+			#user:string
+				Target user (dn format). Ex: "CN=bob,CN=Users,DC=CORP,DC=LOCAL"
+			#group:string
+				Target group (dn format). Ex: "CN=Domain Admins,CN=Users,DC=CORP,DC=LOCAL"
+		"""
+		user = kwargs["user"]
+		group = kwargs["group"]
+
+		if self.engine.add_user_to_group(user, group):
+			info(f"User {user} sucessfully added to {group}")
+		else:
+			error(f"Unable to add {user} to {group}, check privileges or dn")
+
+	def action_remove_from_group(self, kwargs):
+		"""
+		Remove `user` from `group`.
+
+		Arguments:
+			#user:string
+				Target user (dn format). Ex: "CN=bob,CN=Users,DC=CORP,DC=LOCAL"
+			#group:string
+				Target group (dn format). Ex: "CN=Domain Admins,CN=Users,DC=CORP,DC=LOCAL"
+		"""
+		user = kwargs["user"]
+		group = kwargs["group"]
+
+		if self.engine.remove_user_from_group(user, group):
+			info(f"User {user} sucessfully removed from {group}")
+		else:
+			error(f"Unable to remove {user} from {group}, check privileges or dn")
+
 class MSDS_MANAGEDPASSWORD_BLOB(Structure):
 	structure = (
 		('Version','<H'),
@@ -655,6 +869,7 @@ def main():
 	ldap.add_argument("-d", "--domain", required=True, help="The domain as NetBIOS or FQDN")
 	ldap.add_argument("-s", "--ldapserver", required=True, help="The LDAP path (ex : ldap://corp.contoso.com:389)")
 	ldap.add_argument("-b", "--base", default="", help="LDAP base for query (by default, this value is pulled from remote Ldap)")
+	ldap.add_argument("-t", "--type", default="ntlm", choices=['ntlm', 'simple'], help="Authentication type: ntlm (default) or simple")
 	
 	cache.add_argument("-d", "--dir", default=".", type=str, help="Use saved JSON files in specified directory as cache")
 	cache.add_argument("-p", "--prefix", required=True, type=str, help="Prefix of ldeep saved files")
@@ -662,11 +877,16 @@ def main():
 	ntlm = ldap.add_argument_group("NTLM authentication")
 	ntlm.add_argument("-u", "--username", help="The username")
 	ntlm.add_argument("-p", "--password", help="The password used for the authentication")
-	ntlm.add_argument("-H", "--ntlm", help="The NTLM hash used for authentication, ex: aad3b435b51404eeaad3b435b51404ee:a2d4623d306be8e06dbc4e2e8b78353a")
+	ntlm.add_argument("-H", "--ntlm", help="NTLM hashes, format is LMHASH:NTHASH")
 
 	kerberos = ldap.add_argument_group("Kerberos authentication")
 	kerberos.add_argument("-k", "--kerberos", action="store_true", help="For Kerberos authentication, ticket file should be pointed by $KRB5NAME env variable")
 
+	certificate = ldap.add_argument_group("Certificate authentication")
+	certificate.add_argument("--pfx-file", help="PFX file")
+	certificate.add_argument("--cert-pem", help="User certificate")
+	certificate.add_argument("--key-pem", help="User private key")
+
 	anonymous = ldap.add_argument_group("Anonymous authentication")
 	anonymous.add_argument("-a", "--anonymous", action="store_true", help="Perform anonymous binds")
 
@@ -694,17 +914,18 @@ def main():
 			method = "NTLM"
 			if args.kerberos:
 				method = "Kerberos"
-			elif args.username:
-				if args.password:
-					method = "SIMPLE"
-				else:
-					method = "NTLM"
-			elif args.anonymous:
+			elif args.cert_pem or args.pfx_file:
+				method = "Certificate"
+			elif args.anonymous or args.command_ldap == "enum_users":
 				method = "anonymous"
+			elif args.type == "ntlm":
+				method = "NTLM"
+			elif args.type == "simple":
+				method = "SIMPLE"
 			else:
-				error("Lack of authentication options: either Kerberos, Username with Password (can be a NTLM hash) or Anonymous.")
+				error("Lack of authentication options: either Kerberos, Certificate, Username with Password (can be a NTLM hash) or Anonymous.")
 
-			query_engine = LdapActiveDirectoryView(args.ldapserver, args.domain, args.base, args.username, args.password, args.ntlm, method)
+			query_engine = LdapActiveDirectoryView(args.ldapserver, args.domain, args.base, args.username, args.password, args.ntlm, args.pfx_file, args.cert_pem, args.key_pem, method)
 
 			
 		except LdapActiveDirectoryView.ActiveDirectoryLdapException as e:
